package interactive_webcam;


// Non Standard RED- Classes
import java.io.IOException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.red5.server.adapter.ApplicationAdapter;
import org.red5.server.api.IBandwidthConfigure;
import org.red5.server.api.IConnection;
import org.red5.server.api.IScope;
import org.red5.server.api.stream.IServerStream;
import org.red5.server.api.stream.IStreamCapableConnection;
import org.red5.server.api.stream.support.SimpleConnectionBWConfig;

// Lets try and get a bytearray through AMF3
// http://mindprod.com/jgloss/imageio.html#TOBYTES
// http://dl.fancycode.com/red5/api/org/red5/io/amf3/ByteArray.html#writeBytes(byte[])
import org.red5.io.amf3.ByteArray;
import javax.imageio.ImageIO;
import java.awt.image.*;
import java.awt.Transparency;

import java.awt.color.ColorSpace;
import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.File;

// For monitoring of the stream directory 
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.Timer;

public class Manager extends ApplicationAdapter {
	
	// ********  Global Class Objects *********
    // Creating a new instance of the Serial_Port class to avoid static and non-static problems with the
    // serial port methods and functions
       Serial_Port SERIAL = new Serial_Port();
    
	// ********  Global Class DataMembers *********
	// For Function getBMPImage
	private final static int BMP_FILE_HEADER_LEN = 14;  // used for files
    private final static int BMP_INFO_HEADER_LEN = 40;  // typical header len
	
    // For monitoring the streams directory
	private final static String RELATIVE_STREAM_DIRECTORY_PATH = "./dist/webapps/interactive_webcam/streams";
	private boolean StreamsDirModified = false;
	private int InitialStreamsFileAmmount = 0;
	
	// For Red5
	private IScope appScope;

	private IServerStream serverStream;
	
	private static final Log log = LogFactory.getLog(Manager.class);
	
	public boolean appStart(IScope app) {
		appScope = app;
		
		return true;
	}
	
	@Override
	public boolean appConnect(IConnection conn, Object[] params) {
		log.info("Client connected " + conn.getClient().getId() + " conn "
				+ conn);
		// Trigger calling of "onBWDone", required for some FLV players
		measureBandwidth(conn);
		if (conn instanceof IStreamCapableConnection) {
			IStreamCapableConnection streamConn = (IStreamCapableConnection) conn;
			SimpleConnectionBWConfig bwConfig = new SimpleConnectionBWConfig();
			bwConfig.getChannelBandwidth()[IBandwidthConfigure.OVERALL_CHANNEL] =
				1024 * 1024;
			bwConfig.getChannelInitialBurst()[IBandwidthConfigure.OVERALL_CHANNEL] =
				128 * 1024;
			streamConn.setBandwidthConfigure(bwConfig);
		}
	
		log.info("Returning the conn and params: " + getClients().size()); // just a unique number
		
		// Start the timer for checking the streams folder
		StreamCheckTimer.start();
		
		return super.appConnect(conn, params);

	}
	
    @Override
	public void appDisconnect(IConnection conn) {
	 // Stop the timer for checking the streams folder reset values
		StreamCheckTimer.stop();
		StreamsDirModified = false;
	    InitialStreamsFileAmmount = 0;
	
		if (appScope == conn.getScope() && serverStream != null) {
			log.info("Closing server stream using appDisconnect");
			serverStream.close();
		}
		super.appDisconnect(conn);
	}
    
// Non Red-5 Specific Functions
	public double add(double a, double b) {
		log.info("Returning add results");
		return a + b;
	}
	
	// ## NAME: Rotate_Servo
	// ## DATE: 12/16/2007
	// ## AUTHOR: Charles Palen (palen1c@gmail.com)
	// ## DESCRIPTION: This function can be called by a remote swf in order to rotate
	// ## a servo either left or right.  The servo will only rotate a small
	// ## fixed amount for the time being, so the function will need to be called
	// ## numerous times if large ranges are desired.
	// ##
	// ## PRECONDITIONS: Pass a to move left d to move right
	// ## POSTCONDITIONS: NONE	
	public String Rotate_Servo(String _Direction) {
	
        System.out.println("Direction character is: " + _Direction);
	
	 // Opening a connection to the serial port with the speed for the GPS
        boolean S_Connected = SERIAL.Connect("/dev/ttyS1",2400);
        
     // Only try to send commands if the serial port was connected
        if (S_Connected == true) {
        	 try{
                 SERIAL.Send_Data(_Direction);
                 SERIAL.Send_Data(_Direction);
             } catch(IOException e) {log.info("Rotate_Servo: Serial Data Send FAILED-" + e);}
        	
          SERIAL.Disconnect();
        }
        
	 return "Data Sent Over Serial Port";
	}

	// ## NAME: Save_ScreenShot
	// ## DATE: 5/13/2008
	// ## AUTHOR: Charles Palen (palen1c@gmail.com)
	// ## DESCRIPTION: This function can be called by a remote swf in order to save a
	// ## screen capture of an image to the server.  It makes use the bytearray in
	// ## Flex and flash player 9+  
	// ## IMPORTANT:  Its imperitive that the parameter be an object of org.red5.io.amf3.ByteArray
	// ## PRECONDITIONS: 1. Pass a byte array from a flex3 app using netconnection.call(Save_Screen...etc
	// ##                2. Pass the width of the original image as an integer
	// ##                3. Pass the height of the original image as an integer
	// ## POSTCONDITIONS: Trys to get a ByteArray
	public String Save_ScreenShot(ByteArray _RAWBitmapImage, int _ImgWidth, int _ImgHeight) {
	
		// The RAWBitmapImage may be over 1MB in size!
		   int BCurrentlyAvailable = _RAWBitmapImage.bytesAvailable();
		   int BWholeSize = _RAWBitmapImage.length();
		   byte c[] = new byte[BWholeSize];
		   _RAWBitmapImage.readBytes(c);
		
		// Transform the byte array into a buffered image
		//DataBuffer db = new DataBufferByte(c, c.length);
		ByteArrayInputStream db = new ByteArrayInputStream(c);
		//WritableRaster raster = Raster.createInterleavedRaster(db, _ImgWidth, _ImgHeight, 3*_ImgWidth, 3, new int[]{2,1,0}, null);
        //ColorSpace cs = ColorSpace.getInstance(ColorSpace.CS_sRGB);
        //ColorModel cm = new ComponentColorModel(cs, false, false, Transparency.OPAQUE, DataBuffer.TYPE_BYTE);
		
		if(BCurrentlyAvailable > 0) {
             System.out.println("The byte Array currently has " + BCurrentlyAvailable + " bytes. The Buffer has " + db.available());
// DEBUG     

	  
			 //for (int row = 0; row < db.available(); row++) {
 
//                    System.out.println("data is: " + db.read());
//                   }

// END DEBUG	 
			 // Try and read the byte array into a buffered image
			 try{
			     BufferedImage JavaImage = ImageIO.read(db);
			    //BufferedImage JavaImage = new BufferedImage(cm, raster, false, null);

			    // Now lets try and write the buffered image out to a file
				if(JavaImage != null) {
			       ImageIO.write(JavaImage, "PNG", new File("Red5ScreenShot.png"));
				}
			 
			 } catch(IOException e) {log.info("Save_ScreenShot: Writing of screenshot failed " + e); System.out.println("IO Error " + e);}
		}
        
	 return "End of save screen shot";
	}
	
	// ## NAME: Check_Stream_Modification
	// ## DATE: 5/20/2008
	// ## AUTHOR: Charles Palen (palen1c@gmail.com)
	// ## DESCRIPTION: This function checks the streams directory of this application for its modification
	// ##              date to be changed since the application started.  If it is found that the
	// ##              modification date changes(files added), then this will start calling
	// ##              "Rotate_Servo" over and over again with the character 'r'.
	// ##             
	// ##              If a custom external usb device is connected to the computer, 'r' specifies
	// ##              the flashing of a red light to notify the user that new videos are present
	// ##
	// ##              This function fires in conjunction with a timer every 30 seconds or so
	// ## PRECONDITIONS: The timer must be started that corresponds to this actionlistener before this will
	// ##                start checking.
	// ## POSTCONDITIONS: If it is found that the date has changed, this will stop checking the directory and
	// ##                 start sending out 'r' through the serial port until the application is restarted.
	ActionListener Check_Stream_Modification = new ActionListener() {
	   public void actionPerformed(ActionEvent actionEvent) {
		try{
			if(StreamsDirModified == false) {
			   // This app runs as if it is in the root of the red5 dir - pointing to this apps stream directory
			   File f = new File(RELATIVE_STREAM_DIRECTORY_PATH);
			   File[] FileList = f.listFiles();
			   
			   // Set the initial modification time value
			   if ( InitialStreamsFileAmmount == 0) {
				   InitialStreamsFileAmmount = FileList.length;
			   }
			
			   // Start blinking the red light
			   if ( (InitialStreamsFileAmmount != 0) && (InitialStreamsFileAmmount != FileList.length) ) {
				   StreamsDirModified = true;
			   }
			
		       System.out.println(f.getCanonicalPath());
               System.out.println("Number of Files: " + FileList.length);
			} else {
				// Blink the red light to inform an external user of new recordings (send r character over serial)
				   Rotate_Servo("r");
			}
           } catch(IOException e) {log.info("Check_Stream_Modification: Folder mod time failed-" + e);}
	   }
	};

	    // Timer for file modifiction checking
	private Timer StreamCheckTimer = new Timer(30000,Check_Stream_Modification); // Set to 30 seconds
	
	/**
     * Creates an BufferedImage using the data from inStream.
     * If an image cannot be parsed from the data, returns null. 
	 * http://forum.java.sun.com/thread.jspa?threadID=5261905
     */
    private static BufferedImage getBMPImage(InputStream inStream) {
        BufferedImage finalImage;
 
        int imageSize;
        int imageHeight;
        int imageWidth;
        int bitCount;
        int colorsUsed;
 
        byte[] fileHeader = new byte[BMP_FILE_HEADER_LEN];
        byte[] info = new byte[BMP_INFO_HEADER_LEN];
 
        byte[] rgbData;
        int[] imageData;
 
        try {
            inStream.read(fileHeader, 0, fileHeader.length);
            
            // return null if not a bitmap
            if ((fileHeader[0] != 0x42 || fileHeader[1] != 0x4D)
             && (fileHeader[0] != 0x00 || fileHeader[1] != 0x03)) {
                return null;
            }
             
            inStream.read(info, 0, info.length);
 
            // read in parameters from info header
            imageWidth = (int)info[4]&0xFF | ((int)info[5]&0xFF) << 8
             | ((int)info[6]&0xFF) << 16 | ((int)info[7]&0xFF) << 24;
 
            imageHeight = (int)info[8]&0xFF | ((int)info[9]&0xFF) << 8
             | ((int)info[10]&0xFF) << 16 | ((int)info[11]&0xFF) << 24;
 
            bitCount = (int)info[14]&0xFF | ((int)info[15]&0xFF) << 8;
 
            imageSize = (int)info[20]&0xFF | ((int)info[21]&0xFF) << 8
             | ((int)info[22]&0xFF) << 16 | ((int)info[23]&0xFF) << 24;
 
            colorsUsed = (int)info[32]&0xFF | ((int)info[33]&0xFF) << 8
             | ((int)info[34]&0xFF) << 16 | ((int)info[35]&0xFF) << 24;
 
            if (bitCount == 24) { // 24-bit format.
                int padding = imageSize / imageHeight - imageWidth * 3;
                imageData = new int[imageHeight * imageWidth];
                rgbData = new byte[(imageWidth * 3 + padding) * imageHeight];
 
                inStream.read(rgbData, 0, rgbData.length);
 
                int rgbNdx = 0;
                for (int row = 0; row < imageHeight; row++) {
                    for (int col = 0; col < imageWidth; col++) {
                        imageData[imageWidth * (imageHeight - row - 1) + col] =
                         (int)rgbData[rgbNdx++]&0xFF
                         | ((int)rgbData[rgbNdx++]&0xFF) << 8
                         | ((int)rgbData[rgbNdx++]&0xFF) << 16
                         | 0xFF000000;
                    }
 
                    rgbNdx += padding;
                }
 
                finalImage = new BufferedImage(imageWidth, imageHeight, BufferedImage.TYPE_INT_ARGB);
                finalImage.setRGB(0, 0, imageWidth, imageHeight, imageData, 0, imageWidth);
            }
            else if (bitCount == 8) {  // 8-bit format
                int numColors = colorsUsed > 0 ? colorsUsed : 256;
 
                int[] imagePalette = new int[numColors];    // int format
                byte[] inPalette = new byte[numColors * 4]; // argb format
 
                // fixes cases where the image size has not been calculated
                if (imageSize == 0) {
                    imageSize = (((imageWidth * bitCount + 31) & ~31) >> 3)
                     * imageHeight;
                }
                // read in the palette
                inStream.read(inPalette, 0, inPalette.length);
 
                int byteNdx = 0;
                for (int ndx = 0; ndx < numColors; ndx++) {
                    imagePalette[ndx] = (int)inPalette[byteNdx++]&0xFF
                     | ((int)inPalette[byteNdx++]&0xFF) << 8
                     | ((int)inPalette[byteNdx++]&0xFF) << 16
                     | 0xFF000000;
 
                    byteNdx++;
                }
 
                // read in image
                int padding = imageSize / imageHeight - imageWidth;
                imageData = new int[imageWidth * imageHeight];
                rgbData = new byte[(imageWidth + padding) * imageHeight];
 
                inStream.read(rgbData, 0, rgbData.length);
 
                byteNdx = 0;
                for (int row = 0; row < imageHeight; row++) {
                    for (int col = 0; col < imageWidth; col++) {
                        imageData[imageWidth * (imageHeight - row - 1) + col] =
                         imagePalette[rgbData[byteNdx++]&0xFF];
                    }
 
                    byteNdx += padding;
                }
 
                finalImage = new BufferedImage(imageWidth, imageHeight, BufferedImage.TYPE_INT_ARGB);
                finalImage.setRGB(0, 0, imageWidth, imageHeight, imageData, 0, imageWidth);
            }
            else {
                finalImage = null;
            }
 
            inStream.close();
            return finalImage;
        }
        catch (Exception ex) {
            return null;
        }
    }

}
